#!/usr/bin/env python
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-

"""Integration of package managers into SystemUpdater"""

import logging
import os
from gettext import gettext as _
import json
import apt
from SystemUpdater.Core.utils import (
    check_free_space,
    get_broken_details
)
from apt import Cache
from SystemUpdater.Core.UpdateList import LocalUpgradeDataList
import threading
import fcntl
import shutil
from SystemUpdater.Core.utils import emulate_calcul_broken
import subprocess
import threading
from SystemUpdater.Core.DataAcquisition import get_east_8_time

class UpdateEssentialItem():
    def __init__(self):
        self.essential_pkgs = []
        self.remove_white = []

        self.read_path = '/usr/share/kylin-update-desktop-config/config/kylin-update-desktop-system.json'
        self.update()

    def update(self):
        if os.path.exists(self.read_path):
            with open(self.read_path,'r') as f:
                try:
                    data = json.load(f)
                    self.essential_pkgs = data['install_list']
                except Exception as exc:
                    logging.warning(exc)
                try:
                    self.remove_white = data['remove_white_list']
                except Exception as exc:
                    pass
                    # logging.warning(exc)
    
    def check_essential(self,remove_pkgs):
        logging.info("Check: For remove of Essential Packages...")
        for pkg in remove_pkgs:
            if pkg in self.essential_pkgs:
                logging.error("Essential Packages(%s) will be Removed...",pkg)
                return False
        return True
    
    def check_white(self,remove_pkgs):
        logging.info("Check: remove pkg in White Packages...")
        new_upgrade_list = list(set(remove_pkgs) & set(self.remove_white))

        for pkg in new_upgrade_list:
            if pkg in remove_pkgs:
                logging.info("%s will be remove in remove pkg...",pkg)
                remove_pkgs.remove(pkg)

class NowUpgradePara:
    """
    Represent the (potentially partial) results of an unattended-upgrades
    run
    """
    def __init__(self,                  
                upgrade_groups,                 
                upgrade_mode,    
                single_pkgs,
                ):
        #组列表中包含的包
        self.upgrade_groups = upgrade_groups

        #升级的模式
        self.upgrade_mode = upgrade_mode
        #推送的可升级的单包
        self.single_pkgs = single_pkgs

class InstallBackend():
    ACTION_DEFUALT_STATUS = -1
    ACTION_UPDATE = 0
    ACTION_INSTALL = 1
    ACTION_INSTALL_DEB = 2
    ACTION_CHECK_RESOLVER = 3
    ACTION_DOWNLOADONLY = 4
    ACTION_FIX_BROKEN = 5
    ACTION_REMOVE_PACKAGES = 6
    ACTION_FIX_INCOMPLETE = 7
    ACTION_CLEAN = 8
    ACTION_INSTALL_SHUTDOWN = 9
    ACTION_BACKGROUND_UPGRADE = 10
    ACTION_CHECK_BROKEN = 11

    MODE_DEFAULT_STATUS = -1
    
    #1、ACTION_INSTALL 安装的子类
    #部分升级
    MODE_UPGRADE_PARTIAL = 0
    #全部升级
    MODE_UPGRADE_ALL = 1
    #系统全盘升级
    MODE_UPGRADE_SYSTEM = 2
    #后端内部安装包使用 目前 更新配置包和升级本身使用
    MODE_UPGRADE_SINGLE = 3

    #2、更新的子类
    MODE_UPDATE_CACHE = 0
    MODE_UPDATE_ALL = 1

    def __init__(self, window_main, action,mode = MODE_DEFAULT_STATUS):
        self.window_main = window_main
        self.cache = window_main.cache
        self.action = action
        self.action_mode = mode

        #对所有的操作进行枷锁 当执行完毕后返回到_action_done 进行解锁，除了计算依赖模式 不进入_action_done 单独解锁
        self.window_main.is_working = action

        #更新的时候此对象还未生成
        if self.window_main.update_list !=  None:
            #本地维护的升级列表和需要升级的包
            self.upgrade_data = window_main.update_list.local_upgrade_data
        else:
            #生成一个空的列表 防止出现错误 在更新和安装单包的时候这个对象还没生成
            self.upgrade_data = LocalUpgradeDataList({},[],[],[])

        self.update_essential = UpdateEssentialItem()

        #要拿到升级列表必须调用_make_upgrade_list
        self.now_upgrade = NowUpgradePara([],self.MODE_DEFAULT_STATUS,[])

        self.trans_error_details = ''
        
        self.inhibit_shutdown = InhibitShutdownLock()

        #当下载数量大于100个包时  就认为属于大版本升级 开启重试机制
        self.is_version_upgrade = False
        #True表示需要重试安装
        self.is_need_retry = False

    def start(self,_upgrade_mode = MODE_DEFAULT_STATUS,partial_upgrade_list = []):
        _success = True
        header = ''
        desc = ''

        #安装升级包 首先必须调用ACTION_CHECK_RESOLVER 计算依赖解决方便 标记cache 进行升级
        if self.action == self.ACTION_INSTALL:
            self.now_upgrade.upgrade_mode = _upgrade_mode

            if _upgrade_mode != self.MODE_UPGRADE_SINGLE:
                #拿到升级列表
                self.now_upgrade.upgrade_groups,self.now_upgrade.single_pkgs = self._make_groups_list(self.upgrade_data,_upgrade_mode,partial_upgrade_list)

            pkgs_install,pkgs_upgrade,pkgs_remove = self._get_mark_from_cache(self.cache,self.upgrade_data.adjust_pkgs,_upgrade_mode)

            logging.info("INSTALL install:%d , upgrade:%d remove:%d",len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove))
            
            #当下载数量大于200个包时  就认为属于大版本升级 开启重试机制
            if len(pkgs_install) + len(pkgs_upgrade) > 2:
                logging.info("Open a major version upgrade and Retry mechanism on...")
                self.is_version_upgrade = True
            
            #检查是否存在可升级的包
            if len(pkgs_install) == 0 and len(pkgs_upgrade) == 0 and len(pkgs_remove) == 0:
                #FIXME: 没有需要升级和安装的包  检查升级包是否不存在本次升级
                header = _("There is an exception in the update package.")
                desc = _("This update cannot detect the upgradeable package.")
                self._action_done(self.action,False,False,header,desc)
                return

            #检查移除重要软件包
            _success = self.update_essential.check_essential(pkgs_remove)
            if _success == False:
                header = _("There is an exception in the update package.")
                desc = _("You request the removal of a system-essential package.")
                self._action_done(self.action,False,False,header,desc)
                return

            #检查磁盘的状态 
            _success,desc = check_free_space(self.cache)
            if _success == False:
                header = _("Disk space is insufficient, please clean the disk and then upgrade")
                #当单包安装检查内存不够的时候，报错到控制面板 单包安装属于更新才使用的所以直接赋值为更新 再进行之后的报错处理
                if self.now_upgrade.upgrade_mode == self.MODE_UPGRADE_SINGLE:
                    self.action = self.ACTION_UPDATE

                self._action_done(self.action,False,False,header,desc)
                return
                
            logging.info("Disk Check finished...")
            try:
                self.commit(pkgs_install, pkgs_upgrade, pkgs_remove)
            except Exception as e:
                logging.error(e)
        #计算依赖解决方案
        elif self.action == self.ACTION_CHECK_RESOLVER:
            #被删除的包列表
            delete_pkgs = []
            #被删除包的描述
            raw_description = []
            #那些包升级导致删除包 全盘升级不包含 
            delete_desc = []
            
            #检查依赖之前 判断是否需要重启aptdeamon 目前为了生效限速功能
            self.window_main.check_restart_aptdeamon()

            if _upgrade_mode != self.MODE_UPGRADE_SYSTEM:
                #获取要升级的组列表
                self.now_upgrade.upgrade_groups,self.now_upgrade.single_pkgs = self._make_groups_list(self.upgrade_data,_upgrade_mode,partial_upgrade_list)
                #获取要升级和安装的包列表
                pkgs_install,pkgs_upgrade = self._make_pkgs_list(self.cache,self.upgrade_data.upgrade_groups_pkgs,self.now_upgrade.upgrade_groups,self.now_upgrade.single_pkgs)
                #计算解决依赖关系
                _success,delete_pkgs,delete_desc,header,desc = self._make_problem_resolver(self.cache,pkgs_install,pkgs_upgrade)
                pkgs_install,pkgs_upgrade,pkgs_remove = self._get_mark_from_cache(self.cache,self.upgrade_data.adjust_pkgs,_upgrade_mode)
            else:
                # 使用全盘升级 全盘使用dist-upgrade
                try:
                    if self.cache.get_changes():
                        self.cache.clear()
                    _success = self.cache._depcache.upgrade(True)
                    pkgs_install,pkgs_upgrade,pkgs_remove = self._get_mark_from_cache(self.cache,self.upgrade_data.adjust_pkgs,_upgrade_mode)
                except Exception as e:
                    _success = False
                    desc = str(e)
                    logging.error(desc)

            logging.warning("ProblemResolver of the deletion package list:%s",str(pkgs_remove))
            self.update_essential.check_white(pkgs_remove)

            logging.info("RESOLVER install:%d , upgrade:%d remove:%d",len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove))
            is_remove_pkgs = len(pkgs_remove) != 0

            #补充删除包的描述信息,删除描述
            for pkg in pkgs_remove:
                pkg_obj = self.cache[pkg]
                raw_description.append(getattr(pkg_obj.candidate, "summary", ''))
                delete_desc.append('')
            
            #解锁后端状态
            self.window_main.is_working = self.ACTION_DEFUALT_STATUS

            if _upgrade_mode != self.MODE_UPGRADE_SYSTEM:
                self.window_main.dbusController.UpdateDependResloveStatus(_success,is_remove_pkgs,pkgs_remove,raw_description,delete_desc,header,desc)
            else:
                self.window_main.dbusController.DistupgradeDependResloveStatus(_success,is_remove_pkgs,pkgs_remove,raw_description,delete_desc,header,desc)

            if is_remove_pkgs:
                threading_emulate = threading.Thread(target=self._emulate_calcul_delete,args=((pkgs_upgrade + pkgs_install),))
                threading_emulate.start()
        #修复未满足的依赖关系 与apt-get -f install匹配
        elif self.action == self.ACTION_FIX_BROKEN:
            self.fix_broken()
        # 修复不完整的安装dpkg configure -a
        elif self.action == self.ACTION_FIX_INCOMPLETE:
            self.fix_incomplete()
        #卸载包
        elif self.action == self.ACTION_REMOVE_PACKAGES:
            self.purge_packages(partial_upgrade_list)
        #清空所有下载的文件
        elif self.action == self.ACTION_CLEAN:
            self.clean()
        #更新cache
        elif self.action == self.ACTION_UPDATE:
            #检查依赖之前 判断是否需要重启aptdeamon 目前为了生效限速功能
            self.window_main.check_restart_aptdeamon()
            self._backup_current_serverid()
            self.update()

    def start_alone(self,partial_upgrade_list = [],_is_install = False):
        #单独下载功能
        if self.action == self.ACTION_DOWNLOADONLY: 
            try:
                self.download_deb(partial_upgrade_list)
            except Exception as e:
                logging.error(str(e))
        # 安装本地deb包的接口
        elif self.action == self.ACTION_INSTALL_DEB: 
            try:
                self.install_deb(install_path = partial_upgrade_list, install_force = _is_install)
            except Exception as e:
                logging.error(str(e))
        # 安装在线包的接口
        elif self.action == self.ACTION_BACKGROUND_UPGRADE:
            try:
                pkgs_install = [ str(pkg) for pkg in partial_upgrade_list]
                logging.info("Install deb package, open cache")
                cache = Cache()
                self.window_main.sqlite3_server.deb_metadata.update({"background_list":pkgs_install})
                for pkg_name in pkgs_install:
                    if pkg_name not in cache:
                        if "=" not in pkg_name or ("=" in pkg_name and pkg_name.split("=")[0] not in cache):
                        # 没找到包或格式不正确
                            self.window_main.dbusController.UpdateInstallFinished(False, pkgs_install, "'"+pkg_name+"' is not in cache", "")
                            return
                    else:
                        pkg = cache[pkg_name]
                        if pkg.installed:
                            self.window_main.sqlite3_server.deb_metadata['background_list_old_versions'].update({pkg.name:str(pkg.installed.version)})
                        else:
                            self.window_main.sqlite3_server.deb_metadata['background_list_old_versions'].update({pkg.name:str("")})
                self.commit_background(pkgs_install,[],[])
            except Exception as e:
                logging.error(str(e))
                self.window_main.dbusController.UpdateInstallFinished(False, pkgs_install, str(e), "")

    def _backup_current_serverid(self):
        import shutil

        SERVERIDPATH = "/var/lib/kylin-software-properties/config/serverID.conf"
        SERVERIDPATH1 = "/var/lib/kylin-software-properties/config/updateID.conf"
        BACKUPPATH   = "/var/lib/kylin-system-updater/json/"

        try:
            if not os.path.exists(BACKUPPATH):
                os.makedirs(BACKUPPATH, mode=0o755, exist_ok=True)

            if os.path.isfile(SERVERIDPATH):
                dst = shutil.copy(SERVERIDPATH, BACKUPPATH)
                logging.info("Backup serverID conf completed: %s...",dst)
            else:
                logging.info("ServerID conf is not exists...")

            if os.path.isfile(SERVERIDPATH1):
                dst = shutil.copy(SERVERIDPATH1, BACKUPPATH)
                logging.info("Backup updateID conf completed: %s...",dst)
            else:
                logging.info("updateID conf is not exists...")

        except Exception as e:
            logging.error("%s",e)

    def update(self):
        """Run a update to refresh the package list"""
        raise NotImplementedError

    def clean(self):
        """Remove all downloaded files.t"""
        raise NotImplementedError

    def commit(self, pkgs_install, pkgs_upgrade, pkgs_remove,pkgs_purge = []):
        """Commit the cache changes """
        raise NotImplementedError

    def install_deb(self,install_path,install_force):
        """install_deb the cache changes """
        raise NotImplementedError

    def fix_broken(self):
        """fix_broken_depends"""
        raise NotImplementedError

    def fix_incomplete(self):
        """fix_incomplete_install"""
        raise NotImplementedError

    def purge_packages(self,pkgs_purge):
        """purge_packages"""
        raise NotImplementedError

    def _emulate_calcul_delete(self,total_pkg):
        # 多线程的方式来计算删除的软件包
        logging.info("[DELETE_DETAILS] Start calculating delete dependencies...")
        delete_pkgs = []
        last_delete_count = 0
        cache = Cache()

        with cache.actiongroup():
            if cache.get_changes():
                cache.clear()
            for pkg in total_pkg:
                if '#' in pkg:
                    pkg = pkg.split("#")[0]
                elif '=' in pkg:
                    pkg = pkg.split("=")[0]

                try:
                    cache[pkg].mark_install()
                except Exception as e:
                    logging.error(str(e))
                    break

                if cache.delete_count != last_delete_count:
                    last_delete_count = cache.delete_count
                    for pkg_obj in cache.get_changes():
                        if pkg_obj.marked_delete and pkg_obj.name not in delete_pkgs:
                            delete_pkgs.append(pkg_obj.name)
                            delete_pkgs.append(pkg)
                            logging.info("[DELETE_DETAILS] %s Will be deleted Due to Install or Upgrade %s",pkg_obj.name,pkg)
        
        logging.info("[DELETE_DETAILS] End calculating delete dependencies...")
        cache.close()

    #从cache中拿到标记的列表
    def _get_mark_from_cache(self,cache,_adjust_pkgs,upgrade_mode):
        pkgs_install = []
        pkgs_upgrade = []
        pkgs_remove = []

        #全盘升级不做任何的调整
        if upgrade_mode == self.MODE_UPGRADE_SYSTEM:
            adjust_pkgs = []
        else:
            #获取调整包列表 去掉版本号
            adjust_pkgs = [i.split("=")[0] for i in _adjust_pkgs]

        for pkg in cache:
            try:
                if pkg.marked_install:
                    pkgname = pkg.name
                    if pkg.is_auto_installed:
                        pkgname += "#auto"
                    if pkg.name in adjust_pkgs:
                        pkgs_install.append(_adjust_pkgs[adjust_pkgs.index(pkg.name)])
                    else:
                        pkgs_install.append(pkgname)
                elif pkg.marked_upgrade:
                    if pkg.name in adjust_pkgs:
                        pkgs_upgrade.append(_adjust_pkgs[adjust_pkgs.index(pkg.name)])
                    else:
                        pkgs_upgrade.append(pkg.name)
                elif pkg.marked_delete:
                    pkgs_remove.append(pkg.name)
            except KeyError:
                pass
        return pkgs_install,pkgs_upgrade,pkgs_remove

    #获取当前升级的升级列表
    def _make_groups_list(self,upgrade_data,_upgrade_mode,partial_upgrade_list):
        groups_list = []
        pkg_list = []
        
        #部分升级
        if _upgrade_mode == self.MODE_UPGRADE_PARTIAL:
            for elem in partial_upgrade_list:
                #组升级方式
                if elem in upgrade_data.upgrade_groups:
                    groups_list.append(elem)
                #单包升级方式
                elif elem in upgrade_data.single_pkgs:
                    pkg_list.append(elem)
                else:
                    logging.warning("this package(%s) not in selected list",elem)
            
        #全部升级列表
        elif _upgrade_mode == self.MODE_UPGRADE_ALL:
            groups_list = upgrade_data.upgrade_groups
            pkg_list = upgrade_data.single_pkgs
        
        return groups_list,pkg_list

    #从本地中获取本次升级需要升级的包 部分升级和全部升级使用 全盘升级不适用
    def _make_pkgs_list(self,cache,groups_pkgs,groups_list,pkg_list):
        pkgs_install = []
        pkgs_upgrade = []
        
        try:
            #单包的升级方式
            for pkg in pkg_list:
                if cache[pkg].is_installed:
                    pkgs_upgrade.append(pkg)
                else:
                    pkgs_install.append(pkg) 

            #遍历升级组列表
            for group_name in groups_list:
                pkgs_install += groups_pkgs.get(group_name,[]).get('pkgs_install',[])
                pkgs_upgrade += groups_pkgs.get(group_name,[]).get('pkgs_upgrade',[]) 
            
            return pkgs_install,pkgs_upgrade
        except Exception as e:
            logging.error(str(e))
            return [],[]
    
    #计算升级或者安装这个包 是否存在需要卸载某个包
    def _make_delete_pkgs(self,cache,all_delete_pkgs):
        wouldDelete = cache._depcache.del_count
        deleted_pkgs = []
        if wouldDelete > len(all_delete_pkgs):
            for pkg_obj in cache:
                if pkg_obj.marked_delete:
                    if pkg_obj.name in all_delete_pkgs:
                        continue
                    else:
                        deleted_pkgs.append(pkg_obj.name)
            if deleted_pkgs:
                all_delete_pkgs += deleted_pkgs
        return deleted_pkgs

    #将获取本次升级的包 进行计算依赖关系 解决依赖问题
    def _make_problem_resolver(self,cache,pkgs_install = [],pkgs_upgrade = []):
        _success = True
        header = ''
        desc = ''
        #记录出现异常的包
        exception_pkg = ''

        #计算出来的需要删除的包列表
        delete_pkgs = []
        delete_desc = []
        try:
            logging.info("ProblemResolver install:%d , upgrade:%d",len(pkgs_install),len(pkgs_upgrade))
            logging.info("Start calculating dependencies...")
            
            #actiongroup 可以加速计算依赖关系 计算花费的时间将大幅度缩减
            with cache.actiongroup():
                if cache.get_changes():
                    cache.clear()
                resolver = apt.cache.ProblemResolver(cache)

                #标记计算所有需要安装的
                for pkg in pkgs_upgrade + pkgs_install:
                    exception_pkg = pkg

                    pkg_cache = cache[pkg]
                    if pkg_cache.is_upgradable == False and pkg_cache.is_installed == True:
                        logging.warning("pkg(%s) not upgrade or install and will be continue...",pkg)
                        continue

                    #将第二个参数调整为False 当为True时就不能检查缺少依赖的包 默认自动移除掉了 
                    pkg_cache.mark_install(False, True, True)

                    #将可自动升级的标记为自动安装
                    if pkg_cache.is_upgradable == True:
                        auto = pkg_cache.is_auto_installed
                        pkg_cache.mark_auto(auto)

                    if pkg_cache.marked_keep == True:
                        pkg_cache.mark_install(False, False, True)
                        if pkg_cache.marked_keep == True:
                            raise

                    resolver.clear(pkg_cache)
                    resolver.protect(pkg_cache)
                
                resolver.resolve()
            return _success,delete_pkgs,delete_desc,header,desc
        except Exception as e:
            _success = False
            header = _("Could not calculate the upgrade")
            logging.error(header + "(" + exception_pkg + ")" +str(e))
            msg = get_broken_details(cache,False)
            pkg_string = ''
            for pkg in pkgs_install + pkgs_upgrade:
                pkg_string = pkg_string + ' ' + str(pkg)
            logging.info('Resolver calculation Packages List: '+pkg_string+'\n')
            desc ='\n' + msg
            logging.error('\n' + msg)
            threading_emulate = threading.Thread(target=emulate_calcul_broken,args=(pkgs_install + pkgs_upgrade,))
            threading_emulate.start()
            return _success,[],[],header,desc

    #调用aptdeamon结束之后处理的地方 不管是出错还是正常都在此处理
    def _action_done(self, action, is_cancelled,success, error_string,error_desc):
        #后端的状态 到空闲状态
        self.window_main.is_working = self.ACTION_DEFUALT_STATUS
        
        #升级完成后走的分支
        if action == self.ACTION_INSTALL:
            #记录这个过程中是否关机
            self._run_reset_safe()
            if self.now_upgrade.upgrade_mode != self.MODE_UPGRADE_SINGLE:
                self.window_main.configs.setValue("SystemStatus","isabnormalreboot",str(False),True)
            #解锁 禁止关机
            self.inhibit_shutdown.unlock()

            #单包安装模式 单独处理 目前 更新配置包和升级本身使用
            if self.now_upgrade.upgrade_mode == self.MODE_UPGRADE_SINGLE:
                if success:
                    #当单包升级的时候 升级本身时，让程序退出，再重新启动
                    if self.window_main.SELF_PKG_NAME in self.cache and self.cache[self.window_main.SELF_PKG_NAME] in self.cache.get_changes():
                        error_string = _("Update Manager upgrade is complete, please restart the control panel before performing the system update")
                        self.window_main.dbusController.DistUpdateDetectFinished(False,[self.window_main.SELF_PKG_NAME],error_string,error_desc)
                        #升级本身完成后 退出 有systemd 来进行重启服务
                        self.window_main.dbusController.Quit(None)
                    else:
                        #只有安装配置文件包 才会走到此处
                        self.window_main.start_available()
                    return
                else:
                    #当更新过程中安装配置包时 失败时发送错误信号
                    # 防止错误结果太长导致出现控制面板显示太多 目前最多20字符
                    if len(error_string + error_desc) > 20:
                        self.window_main.dbusController.DistUpdateDetectFinished(success,[],error_string,error_desc)
                    else:
                        self.window_main.dbusController.DistUpdateDetectFinished(success,[],error_string+' '+error_desc,error_desc)
                    return
            else:                
                upgrade_content = self.now_upgrade.upgrade_groups + self.now_upgrade.single_pkgs
                
                if self.is_version_upgrade == True and self.is_need_retry == True and success == False:
                    #增加重试次数的限制
                    if self.window_main.retry_limit != 0:
                        logging.warning("Retry the upgrade and installaton(%s) and retry number: %d",upgrade_content,self.window_main.retry_limit)
                        self.window_main.start_install(self.now_upgrade.upgrade_mode,True,upgrade_content)
                        self.window_main.retry_limit = self.window_main.retry_limit - 1
                        return
                    
                self.window_main.sqlite3_server.insert_info(action,self.now_upgrade.single_pkgs,self.now_upgrade.upgrade_groups,[],success,error_string, error_desc)
                self.window_main.dbusController.UpdateInstallFinished(success,upgrade_content,error_string,error_desc)

                P2PInfos = self.window_main.sqlite3_server.get_p2p_info()
                P2PInfos.update({"packageName":str("kylin-system-updater")})
                P2PInfos.update({"source":str("apt-p2p")})
                jsonfile = self.window_main.collector.convertor.dictConvertJson(P2PInfos.copy())
                self.window_main.collector.UpdateMsg("P2PInfos", jsonfile)

                if success:
                    #当组列表为空时 表示现在的单独进行安装某些包或卸载，不发信号到控制面板
                    #升级完成后从升级列表删除
                    for groups in self.now_upgrade.upgrade_groups:
                        self.upgrade_data.upgrade_groups.remove(groups)

                    for pkg in self.now_upgrade.single_pkgs:
                        self.upgrade_data.single_pkgs.remove(pkg)
                    self.window_main.dbusController.CheckRebootRequired("self")
                    #安装完成之后 更新一次cache
                    logging.info("Install or Upgrade successful, so Now update Cache...")
                    self.window_main.only_update_cache()

        elif action == self.ACTION_UPDATE and self.action_mode == self.MODE_UPDATE_ALL:
            if success == False and 'The following signatures' in self.trans_error_details:
                error_desc = _("Check your source public key signature")
            #网络认证错误
            elif success == False and 'does the network require authentication?' in self.trans_error_details:
                error_desc = _("Check if your network requires authentication?")
            
            if success:
                #开始生成列表
                self.window_main.start_available()
            else:
                self.window_main.dbusController.DistUpdateDetectFinished(success,[],error_string+' '+error_desc,error_desc)

            if success == False:
                threading_emulate = threading.Thread(target=self.emulate_update)
                threading_emulate.start()
            self.window_main.sqlite3_server.insert_into_display("check_time",get_east_8_time()[0:-4])

        elif action == self.ACTION_UPDATE and self.action_mode == self.MODE_UPDATE_CACHE:
                self.window_main.dbusController.UpdateDetectFinished(success,[''],error_string,error_desc)

        elif action == self.ACTION_FIX_BROKEN:
            self.window_main.dbusController.FixBrokenStatusChanged(True,success,100,'',error_string,error_desc)
            logging.warning("fix broken packages is complete...")
            
            if success:
                logging.info("Fix broken packages is complete to success...")
                self.window_main.start_available()
            else:
                self.window_main.dbusController.DistUpdateDetectFinished(False,[''],_("Check for update exceptions,fix system APT environment error."),error_string+' '+error_desc)
                logging.warning("fix broken packages is complete to failed...")
            # self.window_main.start_available()
        elif action == self.ACTION_FIX_INCOMPLETE:
            if success:
                logging.info("fix incomplete install success.")
            else:
                logging.warning("fix incomplete install failed.")

        elif action == self.ACTION_REMOVE_PACKAGES:
            self.window_main.dbusController.PurgePackagesFinished(success,error_string,error_desc)
        
        elif action == self.ACTION_INSTALL_DEB:
            import json

            if success == False and 'verification failed' in self.trans_error_details and 'refuse to install' in self.trans_error_details:
                error_string = _("Software package signature verification failed, installation refused.")
                error_desc = ''
            #FIXME: '\r\n: \r\n\r\n'就认为是验证失败
            elif success == False and '\r\n: \r\n\r\n' in self.trans_error_details:
                error_string = _("Package validation failed and installation was rejected.")
                error_desc = ''

            cache = Cache()
            self.window_main.dbusController.InstalldebFinished(success,error_string,error_desc)
            #软件商店/安装器上传
            if str(self.window_main.sqlite3_server.deb_metadata.get('source')) == 'kylin-software-center' \
                or str(self.window_main.sqlite3_server.deb_metadata.get('source')) == 'kylin-installer':
                status = 'success' if (success==True) else 'failed'
                UploadInfos  = {}
                UploadInfos.update({"packageName":str(self.window_main.sqlite3_server.deb_metadata.get('debname'))})
                UploadInfos.update({"source":str(self.window_main.sqlite3_server.deb_metadata.get('source'))})
                UploadInfos.update({"old_version":str(self.window_main.sqlite3_server.deb_metadata.get('old_version'))})
                UploadInfos.update({"new_version":str(self.window_main.sqlite3_server.deb_metadata.get('new_version'))})
                UploadInfos.update({"status":status})
                UploadInfos.update({"errorCode":str(error_string+" "+error_desc)})
                json_file = json.dumps(UploadInfos.copy())
                UploadInfos.pop("source")
                self.window_main.collector.UpdateMsg("UpdateInfos", json_file)

                self.window_main.sqlite3_server.init_metadata()
                UploadInfos  = {}
            
        elif action == self.ACTION_BACKGROUND_UPGRADE:
            import json

            if success == False and 'verification failed' in self.trans_error_details and 'refuse to install' in self.trans_error_details:
                error_string = _("Software package signature verification failed, installation refused.")
                error_desc = ''
            #FIXME: '\r\n: \r\n\r\n'就认为是验证失败
            elif success == False and '\r\n: \r\n\r\n' in self.trans_error_details:
                error_string = _("Package validation failed and installation was rejected.")
                error_desc = ''

            status = 'success' if (success==True) else 'failed'
            errorCode = str(error_string+" "+error_desc)
            UploadInfos  = {}
            upgrade_content = []
            cache = Cache()
            UploadInfos.update({"source":str(self.window_main.sqlite3_server.deb_metadata.get('source'))})
            UploadInfos.update({"status":status})
            UploadInfos.update({"errorCode":errorCode})

            if "background_list" in self.window_main.sqlite3_server.deb_metadata.keys():
                upgrade_content = self.window_main.sqlite3_server.deb_metadata["background_list"]
            
            self.window_main.dbusController.UpdateInstallFinished(success,upgrade_content,error_string,error_desc)

            if not self.window_main.sqlite3_server.check_background_upload(upgrade_content, status, errorCode):
                logging.info("no need upload background upgrade log.")
                return

            for bl in upgrade_content:
                UploadInfos.update({"packageName":str(bl)})
                
                current_version = ''
                if bl in cache:
                    pkg = cache[bl]
                    if pkg.installed:
                        current_version = pkg.installed.version
                UploadInfos.update({"new_version":str(current_version)})

                if 'background_list_old_versions' in self.window_main.sqlite3_server.deb_metadata.keys() and \
                            bl in self.window_main.sqlite3_server.deb_metadata['background_list_old_versions'].keys():
                    tmp = str(self.window_main.sqlite3_server.deb_metadata['background_list_old_versions'].get((bl)))
                    UploadInfos.update({"old_version":str(tmp)})
                else:
                    UploadInfos.update({"old_version":str("unknown")})
                json_file = json.dumps(UploadInfos.copy())
                self.window_main.collector.UpdateMsg("Background", json_file)

    #将安装完成的插入数据库 安装失败的计算那些包安装失败了 分类插入数据库中
    #将安装失败的数据进行返回
    def _make_insert_info(self,success,is_cancelled,_now_upgrade,error_string,error_desc):
        false_num = 0
        try:
            #目前去掉打印 错误日志
            # error_desc = error_desc +'\n\n'+self.trans_error_details
            #在安装失败 and 不是取消升级 则走下面的 否则不计算全部直接插入数据库
            if success == False and is_cancelled == False and self.now_upgrade.upgrade_mode != self.MODE_UPGRADE_SYSTEM:
                #获取当前系统中最新的cache
                fresh_cache = Cache(rootdir=self.window_main.cache.rootdir)
                #获取调整包列表 去掉版本号
                adjust_pkgs = [i.split("=")[0] for i in self.upgrade_data.adjust_pkgs]

                #升级存在单包更新 失败后计算每一个包是否安装完成 相应判断进行插入
                for pkg in _now_upgrade.single_pkgs:
                    pkg_obj = fresh_cache[pkg]
                    
                    #判断是否是源过滤调整的包 调整的话 判断安装版本 来解决是否安装成功
                    if pkg in adjust_pkgs:
                        if pkg_obj.is_installed == True and pkg_obj.is_now_broken == False and  pkg_obj.name + '=' + pkg_obj.installed.source_version in self.upgrade_data.adjust_pkgs:
                            self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[pkg],[],self.upgrade_data.adjust_pkgs,True,'', '')
                        else:
                            false_num += 1
                            logging.warning("This single-package(%s) is not install or upgrade",pkg)
                            self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[pkg],[],self.upgrade_data.adjust_pkgs,False,error_string, error_desc)
                    else:
                    #非调整版本的计算方式
                        if pkg_obj.is_installed == True and pkg_obj.is_upgradable == False and pkg_obj.is_now_broken == False:
                            self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[pkg],[],self.upgrade_data.adjust_pkgs,True,'', '')
                        else:
                            false_num += 1
                            logging.warning("This single-package(%s) is not install or upgrade",pkg)
                            self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[pkg],[],self.upgrade_data.adjust_pkgs,False,error_string, error_desc)
                
                #组的计算是否升级成功的方式 将组按照单包来计算的
                if _now_upgrade.upgrade_groups != []:
                    pkgs_install,pkgs_upgrade = self._make_pkgs_list(self.cache,self.upgrade_data.upgrade_groups_pkgs,_now_upgrade.upgrade_groups,[])
                    group_error_log = ''
                    install_error_pkgs = []
                    total_pkg = pkgs_install + pkgs_upgrade

                    for pkg in pkgs_install + pkgs_upgrade:
                        pkg_obj = fresh_cache[pkg]
                        if pkg in adjust_pkgs:
                            if pkg_obj.is_installed == True and pkg_obj.is_now_broken == False and pkg_obj.installed.source_version + pkg_obj.name in self.upgrade_data.adjust_pkgs:
                                total_pkg.remove(pkg)
                            else:
                                install_error_pkgs.append(pkg)
                        else:
                            if pkg_obj.is_installed == True and pkg_obj.is_upgradable == False and pkg_obj.is_now_broken == False :
                                total_pkg.remove(pkg)
                            else:
                                install_error_pkgs.append(pkg)
                    
                    group_error_log = '\n'+ "This groups-package"+str(install_error_pkgs)+" is not install or upgrade" + '\n'
                    logging.warning(group_error_log)
                    #如果没有
                    if total_pkg == []:
                        self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[],_now_upgrade.upgrade_groups,self.upgrade_data.adjust_pkgs,True,'', '') 
                    else:
                        false_num += 1
                        self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[],_now_upgrade.upgrade_groups,self.upgrade_data.adjust_pkgs,False,error_string,error_desc)
                #关闭临时cache
                fresh_cache.close()
            #其他情况直接写入数据库 不进行计算
            else:
                #取消的话全部是失败
                if is_cancelled == True:
                    false_num = len(_now_upgrade.single_pkgs) + len(_now_upgrade.upgrade_groups)
                self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,_now_upgrade.single_pkgs,_now_upgrade.upgrade_groups,self.upgrade_data.adjust_pkgs,success,error_string, error_desc)
                
            return false_num
        except Exception as e:
            logging.error(str(e))

    def emulate_update(self):
        args = ["apt-get", "update"]

        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        logging.warning(str(p.stdout))
        # return p.stdout

    def _run_shutdown_safe(self):
        logging.info("Run shudown safe manager script before install...")
        args = ["/bin/bash","/usr/share/kysec-maintain/sys-upgrade-pre.sh"]
        try:
            p = subprocess.run(args,timeout=10,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            logging.info(str(p.stdout))
        except Exception as exc:
            logging.error(exc)

    def _run_reset_safe(self):
        logging.info("Run shudown safe manager script after install...")
        args = ["/bin/bash","/usr/share/kysec-maintain/sys-upgrade-post.sh"]
        try:
            p = subprocess.run(args,timeout=10,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            logging.info(str(p.stdout))
        except Exception as exc:
            logging.error(exc)

class InhibitShutdownLock():
    # 禁止关机锁文件路径
    FILELOCK_PATH = "/tmp/lock/"
    SHUTDOWN_BLOCK_FILELOCK = "kylin-update.lock"
    LOGIND_SLEEP_CONF = "/etc/systemd/sleep.conf.d/"
    LOGIND_SLEEP_POSITION = "/etc/systemd/sleep.conf.d/inhibit-sleep.conf"
    SLEEP_CONF_POSITION = "/var/lib/kylin-system-updater/inhibit-sleep.conf"
    
    def __init__(self):
        self.inhibit_lock = None
        self.pidfile = None

    #安装时禁止关机 进行加锁
    def lock(self, caller='kylin-system-updater'):
        """
        Send a dbus signal to logind to not suspend the system, it will be
        released when the return value drops out of scope
        """
        if self.inhibit_lock == None:
            try:
                from gi.repository import Gio, GLib
                logging.info("Start acquiring Shutdown lock(block)...")
                connection = Gio.bus_get_sync(Gio.BusType.SYSTEM)
                
                var, fdlist = connection.call_with_unix_fd_list_sync(
                    'org.freedesktop.login1', '/org/freedesktop/login1',
                    'org.freedesktop.login1.Manager', 'Inhibit',
                    GLib.Variant('(ssss)',
                                ('shutdown:sleep',
                                caller, _("Installing or removing software packages."),
                                'block')),
                    None, 0, -1, None, None)
                self.inhibit_lock = Gio.UnixInputStream(fd=fdlist.steal_fds()[var[0]])
                self.LockedPreventShutdown()
                self._lock_sleep()

                logging.info("Shutdown Has been locked...")
                return True
            except Exception as e:
                logging.error(e)
                return False
        else:
            logging.info("No need to acquired lock and Shutdown Has been locked...")
            return False

    def _lock_sleep(self):
        try:
            if not os.path.exists(self.LOGIND_SLEEP_CONF):
                logging.info("File(%s) is not exists and will be create",self.LOGIND_SLEEP_CONF)
                os.makedirs(self.LOGIND_SLEEP_CONF)
            shutil.copy(self.SLEEP_CONF_POSITION, self.LOGIND_SLEEP_CONF)
            logging.info("File(%s) will be create for forcely inhibit sleep...",self.LOGIND_SLEEP_CONF)
        except Exception as e:
            logging.error(e)
            
    def _unlock_sleep(self):
        try:
            if os.path.exists(self.LOGIND_SLEEP_POSITION):
                os.remove(self.LOGIND_SLEEP_POSITION)
                logging.warning("File(%s) is exists and will be delete and unlock sleep",self.LOGIND_SLEEP_POSITION)
        except Exception as e:
            logging.error(e)
    @staticmethod
    def _check_unlock_sleep():
        try:
            LOGIND_SLEEP_POSITION = "/etc/systemd/sleep.conf.d/inhibit-sleep.conf"
            if os.path.exists(LOGIND_SLEEP_POSITION):
                os.remove(LOGIND_SLEEP_POSITION)
                logging.warning("File(%s) is exists and will be delete and unlock sleep",LOGIND_SLEEP_POSITION)
        except Exception as e:
            logging.error(e)

    #解锁禁止关机 
    def unlock(self):
        try:
            self.unLockedEnableShutdown()
            self._unlock_sleep()
            if self.inhibit_lock != None:
                self.inhibit_lock.close()
                self.inhibit_lock = None
                logging.info("Shutdown Has been unlocked...")
            else:
                logging.info("Not locked and Quitting ...")
        except Exception as e:
            self.inhibit_lock = None
            logging.error("unlock failed." + str(e))

    #安装时禁止关机 进行加锁
    def LockedPreventShutdown(self):
        #不为空是表示以及被锁
        if self.pidfile != None:
            logging.error("self.pidfile file disc not is None,Has been locked...")
            return False

        if not os.path.exists(self.FILELOCK_PATH):
            #不存在创建
            logging.info("File(%s) is not exists and will be create",self.FILELOCK_PATH)
            os.makedirs(self.FILELOCK_PATH)
        else:
            #当目录存在时进行删除 不删除进行创建文件的话会报错 
            # file cannot be locked.[Errno 11] Resource temporarily unavailable
            # 资源被占用报错
            shutil.rmtree(self.FILELOCK_PATH)
            logging.info("File(%s) is exists and will be delete and create",self.FILELOCK_PATH)
            os.makedirs(self.FILELOCK_PATH)        

        try:
            self.pidfile = open(os.path.join(self.FILELOCK_PATH, self.SHUTDOWN_BLOCK_FILELOCK), "w+")
            fcntl.flock(self.pidfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
            return True
        except Exception as e:
            logging.error("file cannot be locked." + str(e))
            self.pidfile = None
            return False

    #解锁禁止关机 
    def unLockedEnableShutdown(self):
        #未加锁退出
        if not self.pidfile:
            logging.info("Not locked and Quitting ...")
            return False
        try:
            fcntl.flock(self.pidfile, fcntl.LOCK_UN)
            self.pidfile.close()
            self.pidfile = None

            # Fix 修复权限问题 当普通用户无法使用 所以直接删除目录
            if os.path.exists(self.FILELOCK_PATH):
                shutil.rmtree(self.FILELOCK_PATH)
                logging.info('Emptying the lockPath(%s) is complete...',self.FILELOCK_PATH)
            else:
                logging.info("Emptying the lockPath(%s) is Failed...",self.FILELOCK_PATH)

            return True
        except Exception as e:
            logging.error("unlock failed." + str(e))
            self.pidfile = None
            return False   


# try aptdaemon
if os.path.exists("/usr/sbin/aptd") \
   and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ:
    # check if the gtkwidgets are installed as well
    try:
        from .InstallBackendAptdaemon import InstallBackendAptdaemon
    except ImportError:
        logging.exception("importing aptdaemon")

def get_backend(*args, **kwargs):
    """Select and return a package manager backend."""
    # try aptdaemon
    if (os.path.exists("/usr/sbin/aptd")
            and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ):
        # check if the gtkwidgets are installed as well
        try:
            return InstallBackendAptdaemon(*args, **kwargs)
        except NameError:
            logging.exception("using aptdaemon failed")

    # nothing found, raise
    raise Exception("No working backend found, please try installing "
                    "aptdaemon or synaptic")
